<?php
namespace FASTPIXEL;
defined('ABSPATH') || exit;
use FASTPIXEL\FASTPIXEL_Config_Model;

if (!class_exists('FASTPIXEL\FASTPIXEL_Cache')) {
    class FASTPIXEL_Cache 
    {
        protected static $instance;
        protected $config;
        protected $functions;
        protected $cache_dir;
        protected $url;
        protected $url_path;
        protected $html_path;
        protected $json_path;
        protected $header_path;
        protected $meta_path;
        protected $request_wait_time = 90;
        protected $debug_mode = false;
        protected $excluded_file_extensions = [
            //images
            'svg',
            'png',
            'ico',
            'jpeg',
            'jpg',
            'bmp',
            'gif',
            //documents
            'txt',
            'doc',
            'xml',
            'xsl',
            //fonts
            'woff',
            'woff2',
            'otf',
            'ttf',
            'eof'
        ];
        protected $buffer;

        function __construct() {
            self::$instance  = $this;
            $this->functions = FASTPIXEL_Functions::get_instance();
            $this->config    = FASTPIXEL_Config_Model::get_instance();

            //loading modules (need them in advanced-cache.php)
            $this->load_modules();

            //define FASTPIXEL_REST_URL for backend actions, using hook because during include rest_url throw error
            add_action('init', function () {
                if (!defined('FASTPIXEL_REST_URL') && function_exists('rest_url')) {
                    define('FASTPIXEL_REST_URL', get_rest_url(get_current_blog_id(), 'fastpixel-website-accelerator' . '/v1/update'));
                }
            });
            if (!defined('FASTPIXEL_API_URL')) {
                define('FASTPIXEL_API_URL', FASTPIXEL_API_HOST . '/api/v1/enqueue');
            }

            do_action('fastpixel_early_init');

            if (!$this->check_request_agent()) {
                return;
            }
            if (!$this->check_request()) {
                return;
            }

            $this->cache_dir = $this->functions->get_cache_dir();
            $this->url = new FASTPIXEL_Url();
            $this->url_path = $this->url->get_url_path();
            $this->html_path = $this->cache_dir . DIRECTORY_SEPARATOR . $this->url_path . 'index.html';
            $this->json_path = $this->cache_dir . DIRECTORY_SEPARATOR . $this->url_path . 'index.json';
            $this->header_path = $this->cache_dir . DIRECTORY_SEPARATOR . $this->url_path . 'headers.json';
            $this->meta_path = $this->cache_dir . DIRECTORY_SEPARATOR . $this->url_path . 'meta';

            //if no cache files exists registering function for cache request
            register_shutdown_function([$this, 'request_page_cache']);
        }

        public static function get_instance() {
            if (empty(self::$instance)) {
                self::$instance = new FASTPIXEL_Cache();
            }
            return self::$instance;
        }
        /**
        * https://stackoverflow.com/questions/38546354/woocommerce-cookies-and-sessions-get-the-current-products-in-cart
        * also confirmed by aguidrevitch at https://outletdepanales.com/ on 23th Jan 2024
        */
        function check_has_items_in_cart() {
            if (!empty($_COOKIE['woocommerce_items_in_cart']) && intval($_COOKIE['woocommerce_items_in_cart']) > 0) {
                return true;
            }
            return false;
        }
        function check_is_logged_in()
        {
            $login_cookies = array(defined('FASTPIXEL_LOGGED_IN_COOKIE') ? FASTPIXEL_LOGGED_IN_COOKIE : (defined('LOGGED_IN_COOKIE') ? LOGGED_IN_COOKIE : ''));
            foreach ($login_cookies as $l_cookie) {
                if (!empty($_COOKIE[$l_cookie])) {
                    return true;
                }
            }
            $logged = false;
            foreach ($_COOKIE as $key => $value) {
                if (strpos($key, "wordpress_logged_in_") !== false) {
                    $logged = true;
                }
            }
            return $logged;
        }

        function check_request_agent() {
            /** 
            * To avoid loops, otherwise request from service will also trigger it
            */
            if (isset($_SERVER['HTTP_USER_AGENT']) && !empty($_SERVER['HTTP_USER_AGENT']) && 
            (strpos($_SERVER['HTTP_USER_AGENT'], 'FastPixel') !== false)
            ) {
                //need to check if this is required
                return false;
            }
            return true;
        }

        function check_request() {
            /**
            * if fastpixeldebug query is present set it for later usage
            */
            if (isset($_SERVER['QUERY_STRING']) &&
                strpos($_SERVER['QUERY_STRING'], 'fastpixeldebug') !== false) {
                $this->debug_mode = true;
            }
            /**
            * Don't ever cache non-GET requests.
            */
            if (!isset($_SERVER['REQUEST_METHOD']) || 'GET' !== $_SERVER['REQUEST_METHOD']) {
                FASTPIXEL_DEBUG::log('CHECKING REQUEST: Not GET request, stopped');
                return false;
            }
            /**
            * No need to activate this class in wordpress cli or cron or ajax request 
            */
            if (defined('WP_CLI') || 
                defined('DOING_CRON') || 
                defined('DOING_AJAX') ||
                (!empty($_SERVER['HTTP_X_REQUESTED_WITH']) && 
                strtolower($_SERVER['HTTP_X_REQUESTED_WITH']) == 'xmlhttprequest')
            ) {
                return false;
            }
            /**
            * getting request url
            */
            $requested_url = new FASTPIXEL_Url();
            /**
            * no need to activate class on api request
            */
            // phpcs:ignore WordPress.Security.NonceVerification.Recommended -- none available before WordPress is loaded.
            if (isset($_GET['rest_route']) || strpos($requested_url->get_url(), 'wp-json') > 0) { //phpcs:ignore
                return false;
            }
            /**
            * run only for index.php
            */
            if (!in_array($_SERVER['SCRIPT_NAME'], ['/index.php'])) {
                return false;
            }
            /**
            * non-existing files are redirected to wordpress index.php, no need to serve them
            */
            $file_extension = pathinfo(basename($requested_url->get_url()), PATHINFO_EXTENSION);
            if (!empty($file_extension) && (in_array($file_extension, $this->excluded_file_extensions))) {
                return false;
            }
            /**
            * do not serve cached pages for users with non empty shopping cart
            */
            if ($this->check_has_items_in_cart()) {
                return false;
            }
            /**
            * do not serve cached pages for logged in users if enabled
            */
            if ($this->check_is_logged_in() && !$this->config->get_option('fastpixel_display_cached_for_logged')) {
                return false;
            }
            /*
            * don't ever cache Admin requests
            */
            if(is_admin()) {
                return false;
            }
            /**
            * don't cache request with nocache query
            */
            if(isset($_SERVER['QUERY_STRING']) && strpos($_SERVER['QUERY_STRING'], 'fastpixeldisable') !== false) {
                return false;
            }
            /**
            * checking excluded urls
            */
            $default_url_exclusions = [
                '/robots.txt',
                '/.htaccess',
                '/favicon.ico',
            ];
            $user_url_exclusions = $this->config->get_option('fastpixel_exclusions');
            if (is_string($user_url_exclusions)) {
                $url_exclusions = explode(' ', $user_url_exclusions);
            } else if (is_array($user_url_exclusions)) {
                $url_exclusions = $user_url_exclusions;
            } else {
                $url_exclusions = [];
            }
            if (!empty($url_exclusions) && is_array($url_exclusions)) {
                $default_url_exclusions = array_merge($default_url_exclusions, $url_exclusions);
            }
            $excluded_urls = array_filter(apply_filters('fastpixel_exclude_url', $default_url_exclusions));
            foreach ($excluded_urls as $excluded_url) {
                // do we want a regular expression match instead?
                // lets start with something easy
                if (!empty(trim($excluded_url))) {
                    $prepared_req_url = $requested_url->get_path() != '/' ? strtolower(rtrim($requested_url->get_path(), '/')) : '/';
                    $prepared_exc_url = $excluded_url != '/' ? strtolower(rtrim(trim($excluded_url), '/')) : '/';
                    if (!empty($prepared_req_url) && $prepared_req_url == $prepared_exc_url) {
                        //need to delete if url is excluded
                        $this->functions->delete_cached_files($requested_url->get_url_path());
                        return false;
                    }
                }
            }
            /**
            * checking excluded urls by params
            */
            $excluded_params = $this->config->get_option('fastpixel_params_exclusions');
            if (is_string($excluded_params)) {
                $excluded_params_exploded = array_filter(explode(" ", $excluded_params));
            } else if (is_array($excluded_params)) {
                $excluded_params_exploded = $excluded_params;
            } else {
                $excluded_params_exploded = [];
            }
            $excluded_params = [
                'action'              => 'elementor', // Elementor
                'elementor-preview'   => '', // Elementor
                'bricks'              => '', // Bricks plugin
                'brizy-edit-iframe'   => '', // Brizy plugin
                'builder'             => '', // Fusion Builder
                'ct_builder'          => '', // Oxygen
                'et_fb'               => '', // Divi
                'fb-edit'             => '', // Fusion Builder
                'fl_builder'          => '', // Beaver Builder
                'preview'             => '', // Blockeditor & Gutenberg
                'tb-preview'          => '', // Themify
                'tve'                 => '', // Thrive
                'uxb_iframe'          => '', // Flatsome UX Builder
                'vc_action'           => '', // WP Bakery
                'vc_editable'         => '', // WP Bakery
                'vcv-action'          => '', // WP Bakery
                'wyp_mode'            => '', // Yellowpencil plugin
                'wyp_page_type'       => '', // Yellowpencil plugin
                'zionbuilder-preview' => '', // Zion Builder plugin
                '_wp-find-template'   => '' //default wp/gutenberg function
            ];
            if(!empty($excluded_params_exploded) && is_array($excluded_params_exploded)) {
                //generating array of excluded params
                foreach($excluded_params_exploded as $param) {
                    if(!empty(trim($param))) {
                        parse_str($param, $parsed_param);
                        $key = strtolower(trim(key($parsed_param)));
                        $excluded_params[$key] = strtolower(trim($parsed_param[$key]));
                    }
                }
            }
            //removing empty items
            $excluded_params = array_filter($excluded_params);
            $excluded_params_keys = array_keys($excluded_params);
            if(!empty($requested_url->get_query())) {
                //parsing request params
                parse_str($requested_url->get_query(), $parsed_request_params);
                //checking url params with excluded keys
                foreach($parsed_request_params as $key => $value) {
                    //check if key is present in excluded array
                    if(in_array($key, $excluded_params_keys)) {
                        //check for excluded value,
                        //if value is present and it matches then set exclusion to true, otherwise set exclusion to true
                        if(!empty($excluded_params[$key])) {
                            if(strtolower($excluded_params[$key]) == strtolower($value)) {
                                //need to delete if url is excluded
                                $this->functions->delete_cached_files($requested_url->get_url_path());
                                return false;
                            } else {
                                continue;
                            }
                        } else {
                            //need to delete if url is excluded
                            $this->functions->delete_cached_files($requested_url->get_url_path());
                            return false;
                        }
                    }
                }
            }
            return true;
        }

        public function return_cached_if_exists() {
            //no need to continue if url_path is empty
            if (empty($this->url_path)) {
                return;
            }
            FASTPIXEL_DEBUG::log('CHECKING FOR CACHED PAGE', $this->url_path);
            $status = $this->functions->check_post_cache_status($this->url->get_url());
            
            //need to check parent for pagination 
            if (preg_match('/page\/\d+/', $this->url_path)) {
                $url_path = preg_replace('/page.+$/i', '', $this->url_path);
                $parent_status = $this->functions->check_post_cache_status($url_path);
                if ($parent_status['local_invalidation_time'] > $status['local_invalidation_time']) {
                    $status['need_cache'] = true;
                    $status['local_invalidation_time'] = $parent_status['local_invalidation_time'];
                }
            }
            //extra check for pages with params
            if (strpos($this->url->get_url(), '?') > 0) {
                $params_url = str_replace('?' . $this->url->get_query(), '', $this->url->get_url());
                $parent_status = $this->functions->check_post_cache_status($params_url);
                if ($status['have_cache'] && $parent_status['local_invalidation_time'] > $status['local_invalidation_time']) {
                    $status['need_cache'] = true;
                    $status['local_invalidation_time'] = $parent_status['local_invalidation_time'];
                }
            } 

            FASTPIXEL_DEBUG::log('CHECKING FOR CACHED PAGE : CACHE DATA', $status );
            //serve_stale is used to return old cached version while new cache was requested
            //if serve_stale is disabled and we have "invalidated" file then return "wordpress page"
            //else serve cached version if exists
            FASTPIXEL_DEBUG::log('CHECKING FOR CACHED PAGE : STALE CACHE Enabled', (bool)$this->config->get_option('fastpixel_serve_stale'));

            if (!$status['have_cache']) {
                FASTPIXEL_Debug::log('No cache files exist, nothing to return, skipping');
                return;
            }
            //if serve stale is enabled
            if ($this->config->get_option('fastpixel_serve_stale')) {
                //if page require new cache
                if ($status['need_cache']) {
                    //checking page creation time and compare with request time
                    $invalidation_time = ($status['global_invalidation_time'] > $status['local_invalidation_time'] ? $status['global_invalidation_time'] : $status['local_invalidation_time']);
                    //if page is not refreshed then we need to request page cache
                    if ($status['html_created_time'] < $invalidation_time) {
                        //doing request if it wasn't done, or >=5 mins left from last request
                        if (!$status['last_cache_request_time'] || ((time() + $this->request_wait_time) > $status['last_cache_request_time'])) {
                            //define constant
                            define('FASTPIXEL_SERVE_STALE_REQUEST', true);
                        }
                    }
                }
            } else 
            //no need to return cached page if serve_stale is disabled and page cache is reset
            if ($status['need_cache']) {
                FASTPIXEL_Debug::log('Cached page is not generated yet, nothing to return, skipping');
                return;
            }

            if (@file_exists($this->html_path) && @is_readable($this->html_path)) {
                $path = $this->html_path;
            } elseif (@file_exists($this->json_path) && @is_readable($this->json_path)) {
                $path = $this->json_path;
                header('Content-Type: application/json; charset=UTF-8');
            }

            if (@file_exists($path) && @is_readable($path)) {
                $modified_time = (int) @filemtime($path);
                if (!empty($modified_time) && !empty($_SERVER['HTTP_IF_MODIFIED_SINCE']) && strtotime($_SERVER['HTTP_IF_MODIFIED_SINCE']) === $modified_time) {
                    $valid_protocols = ['HTTP/1.0', 'HTTP/1.1', 'HTTP/2.0', 'HTTP/3'];
                    $protocol = strtoupper($_SERVER['SERVER_PROTOCOL']) ?? false; 
                    if (!in_array($protocol, $valid_protocols)) {
                        $protocol = $valid_protocols[1]; // use HTTP/1.1 as default protocol
                    }
                    header($protocol . '304 Not Modified', true, 304);
                    exit;
                }
                if (@file_exists($this->header_path) && @is_readable($this->header_path)) {
                    /*
                     * can't use here wordpress native functions(WP_Filesystem) because this function fires early in advanced-cache.php
                     */
                    // phpcs:ignore WordPress.WP.AlternativeFunctions.file_get_contents_file_get_contents -- none available before WordPress is loaded..
                    $headers = json_decode(@file_get_contents($this->header_path)) ?: []; //phpcs:ignore
                    foreach ($headers as $header) {
                        header($header);
                    }
                }
                FASTPIXEL_DEBUG::log('CHECKING FOR CACHED PAGE : RETURNING CACHED PAGE');
                $headers = ['X-FastPixel-Cache: HIT'];
                $mtime = filemtime($path);
                $headers[] = 'Age: ' . (time() - $mtime);
                /*
                 * can't use here wordpress native functions(WP_Filesystem) because this function fires early in advanced-cache.php
                 * reading .html file previously generated by our service
                 */
                // phpcs:ignore WordPress.WP.AlternativeFunctions.file_get_contents_file_get_contents -- none available before WordPress is loaded..
                $buffer = file_get_contents($path); //phpcs:ignore
                $headers = apply_filters('fastpixel_return_cached_page_headers', $headers);
                foreach ($headers as $header) {
                    header($header);
                }
                $js_optimize = $this->config->get_option('fastpixel_javascript_optimization');
                /* blocker */
                if ($this->debug_mode) {
                    // TODO: add debug code
                    if ($js_optimize != 3) {
                        //TODO: replace code with debug
                    }
                }
                $buffer .= '<!-- Optimized and served by FastPixel.io -->';
                //checking if we need to serve cached page for logged user and if user is logged
                if ($this->config->get_option('fastpixel_display_cached_for_logged') && $this->check_is_logged_in()) {
                    //saving buffer
                    $this->buffer = $buffer;
                    //registering hook that will display cached page before page rendering should begin
                    add_action('template_redirect', [$this, 'display_cached_for_logged']);
                } else {
                    if ($this->functions->valid_html($buffer)) {
                        /*
                         * displaying .html file, generated by our service 
                         * no need to escape output, escape functions are not available in advanced-cache.php
                         */
                        // phpcs:ignore WordPress.Security.EscapeOutput.OutputNotEscaped -- none available before WordPress is loaded and no need to escape.
                        echo $buffer; // phpcs:ignore
                        define('FASTPIXEL_STOP_REQUEST', true);
                        @ob_end_flush();
                        flush();
                        exit();
                    }
                }
            }
        }

        public function request_cache_when_serve_stale() {
            FASTPIXEL_Debug::log('REQUEST CACHE WHEN SERVE STALE Started');
            // Define WP_LANG_DIR if not set.
            wp_set_lang_dir();

            // Load early WordPress files.
            if (file_exists(ABSPATH . WPINC . '/compat.php')) {
                require_once ABSPATH . WPINC . '/compat.php';
            }
            require_once ABSPATH . WPINC . '/class-wp-list-util.php';
            require_once ABSPATH . WPINC . '/formatting.php';
            require_once ABSPATH . WPINC . '/meta.php';
            require_once ABSPATH . WPINC . '/functions.php';
            require_once ABSPATH . WPINC . '/class-wp-meta-query.php';
            require_once ABSPATH . WPINC . '/class-wp-matchesmapregex.php';
            require_once ABSPATH . WPINC . '/class-wp.php';
            require_once ABSPATH . WPINC . '/class-wp-error.php';
            require_once ABSPATH . WPINC . '/pomo/mo.php';

            global $wpdb, $table_prefix;
            // Include the wpdb class and, if present, a db.php database drop-in.
            require_wp_db();

            // Set the database table prefix and the format specifiers for database table columns.
            // phpcs:ignore WordPress.WP.GlobalVariablesOverride.Prohibited -- code copied from wordpress core
            $GLOBALS['table_prefix'] = $table_prefix;
            wp_set_wpdb_vars();

            // Start the WordPress object cache, or an external object cache if the drop-in is present.
            wp_start_object_cache();

            // Attach the default filters.
            require_once ABSPATH . WPINC . '/default-filters.php';

            // Initialize multisite if enabled.
            if (is_multisite()) {
                require_once ABSPATH . WPINC . '/class-wp-site-query.php';
                require_once ABSPATH . WPINC . '/class-wp-network-query.php';
                require_once ABSPATH . WPINC . '/ms-blogs.php';
                require_once ABSPATH . WPINC . '/ms-settings.php';
            }

            //loading required libraries
            require_once ABSPATH . WPINC . '/general-template.php';
            require_once ABSPATH . WPINC . '/link-template.php';

            //requiring wp http api classes
            require_once ABSPATH . WPINC . '/http.php';
            //checking if file exists in current version
            if (file_exists(ABSPATH . WPINC . '/class-wp-http.php')) {
                require_once ABSPATH . WPINC . '/class-wp-http.php';
            }
            //added support for older versions (before v6)
            elseif (file_exists(ABSPATH . WPINC . '/class-http.php')) {
                require_once ABSPATH . WPINC . '/class-http.php';
            }
            require_once ABSPATH . WPINC . '/class-wp-http-streams.php';
            require_once ABSPATH . WPINC . '/class-wp-http-curl.php';
            require_once ABSPATH . WPINC . '/class-wp-http-proxy.php';
            require_once ABSPATH . WPINC . '/class-wp-http-cookie.php';
            require_once ABSPATH . WPINC . '/class-wp-http-encoding.php';
            require_once ABSPATH . WPINC . '/class-wp-http-response.php';
            require_once ABSPATH . WPINC . '/class-wp-http-requests-response.php';
            require_once ABSPATH . WPINC . '/class-wp-http-requests-hooks.php';

            // Load multisite-specific files.
            if (is_multisite()) {
                require_once ABSPATH . WPINC . '/ms-functions.php';
                require_once ABSPATH . WPINC . '/ms-default-filters.php';
                require_once ABSPATH . WPINC . '/ms-deprecated.php';
            }
            // Define constants that rely on the API to obtain the default value.
            // Define must-use plugin directory constants, which may be overridden in the sunrise.php drop-in.
            wp_plugin_directory_constants();

            // phpcs:ignore WordPress.WP.GlobalVariablesOverride.Prohibited -- code copied from wordpress core
            $GLOBALS['wp_plugin_paths'] = array();

            //running request
            require_once(FASTPIXEL_PLUGIN_DIR . DIRECTORY_SEPARATOR . 'fastpixel.php');

            //running basic validations
            if (!$this->common_basic_validations_before_cache_request()) {
                exit;
            }
            FASTPIXEL_Debug::log('REQUEST CACHE WHEN SERVE STALE: URL that should be cached', $this->url->get_url());
            if ($this->url) {
                if (!$this->common_excludes_and_meta_validation_before_cache_request($this->url->get_url())) {
                    exit;
                }
                if ($res = $this->cache_request($this->url->get_url())) {
                    FASTPIXEL_Debug::log('REQUEST CACHE WHEN SERVE STALE: Cache Request result', $res);
                    $this->functions->update_post_cache($this->url->get_url_path(), false, true);
                }
            }
            FASTPIXEL_Debug::log('REQUEST CACHE WHEN SERVE STALE Ended');
            exit;
        }

        //function runs on register_shutdown_function
        function request_page_cache($post_id = null)
        {
            $cache_lifetime = $this->config->get_option('fastpixel_cache_lifetime');
            //checking cache lifetime
            if ($cache_lifetime && is_numeric($cache_lifetime) && in_array($cache_lifetime, array(2, 3)) && !empty($this->url) && is_object($this->url)) {
                $cache_status = $this->functions->check_post_cache_status($this->url->get_url());
                //checking 24H
                if ($cache_lifetime == 2 && ($cache_status['have_cache'] && !$cache_status['need_cache']) && time() > $cache_status['html_created_time'] + 24 * 3600) {
                    $this->functions->update_post_cache($this->url_path, true);
                } else 
                //checking 12H
                if ($cache_lifetime == 3 && ($cache_status['have_cache'] && !$cache_status['need_cache']) && time() > $cache_status['html_created_time'] + 12 * 3600) {
                    $this->functions->update_post_cache($this->url_path, true);
                }
            }
            //can't request pages that are displayed for logged in users
            if (function_exists('is_user_logged_in') && is_user_logged_in() && $post_id == null) {
                FASTPIXEL_Debug::log('stopping request for logged in user');
                return false;
            }
            //running validation and request when serve stale, moved here to avoid multiple shutdown callbacks and not duplicating functionality
            if(defined('FASTPIXEL_SERVE_STALE_REQUEST') && FASTPIXEL_SERVE_STALE_REQUEST) {
                $this->request_cache_when_serve_stale();
                //no need to continue execution, stopping after request
                return false;
            }
            //stopping reques tafter cached page displayed
            if (defined('FASTPIXEL_STOP_REQUEST') && FASTPIXEL_STOP_REQUEST) {
                return false;
            }

            //running basic validations
            if (!$this->common_basic_validations_before_cache_request()) {
                return false;
            }
            global $post;
            // Don't cache search, 404, or password protected.
            if (is_404() || is_search() || (!empty($post) && !empty($post->post_password)) || is_feed()) {
                FASTPIXEL_Debug::log('REQUESTING PAGE CACHE Stopped, is 404 or search or feed or password protected page');
                return false;
            }
            // Extra check for rest request, don't run cache on API request
            if (defined('REST_REQUEST') && REST_REQUEST && !defined('FASTPIXEL_SAVE_POST'))  {
                return false;
            }
            //Do not cache not published posts
            if (is_single() && $post->post_status != 'publish') {
                return false;
            }
            $excluded_post_types = apply_filters('fastpixel_excluded_post_types', []);
            if (!empty($post_id) && is_numeric($post_id)) {
                $post_type = get_post_type($post_id);
                if (in_array($post_type, $excluded_post_types)) {
                    FASTPIXEL_Debug::log('REQUESTING PAGE CACHE Stopped, post type is excluded from cache');
                    return false;
                }
            }
            //getting url info
            //check for is_front_page() || is_home() have higher priority
            if (!is_admin() && (is_front_page() || is_home())) {
                $request_url = $this->url->get_url();
                $request_path = $this->url->get_url_path();
            } else if (isset($post_id) && !is_null($post_id) && !empty($post_id)) {
                $url = new FASTPIXEL_Url($post_id);
                $request_url = $url->get_url();
                $request_path = $url->get_url_path();
            } else if (isset($post->ID) && !empty($post->ID)) {
                $url = new FASTPIXEL_Url($post->ID);
                $request_url = $url->get_url();
                $request_path = $url->get_url_path();
                $initial_request_url = $this->url->get_url();
                $initial_request_path = $this->url->get_url_path();
                //using request url if post url match request url
                if (strpos($initial_request_url, $request_url) >= 0) {
                    $request_url = $initial_request_url;
                    $request_path = $initial_request_path;
                }
            } else {
                $request_url = $this->url->get_url();
                $request_path = $this->url->get_url_path();
            }
            //check for pages that should be excluded from cache
            if (!$this->common_excludes_and_meta_validation_before_cache_request($request_url)) {
                return false;
            }
            //Doing Page Cache request(before page renders)
            if ($this->cache_request($request_url)) {
                $this->functions->update_post_cache($request_path, false, true);
                FASTPIXEL_Debug::log('REQUESTING PAGE CACHE Ended Successfully');
                return true;
            } else {
                FASTPIXEL_Debug::log('REQUESTING PAGE CACHE Ended with error');
                return false;
            }
        }

        protected function common_basic_validations_before_cache_request() {
            if ($this->debug_mode) {
                FASTPIXEL_Debug::log('Common Validation: Stopped. Debug request.');
                return false;
            }
            $diag = FASTPIXEL_Diag::get_instance();
            if (!$diag->run_activation_tests()) {
                FASTPIXEL_Debug::log('Common Validation: Stopped. Diagnostics errors');
                return false;
            }
            return true;
        }

        protected function common_excludes_and_meta_validation_before_cache_request($input_url = false) {
            if (!$input_url) {
                return false;
            }
            $url = new FASTPIXEL_Url(null, $input_url);
            //check for pages that should be excluded from cache
            $excludes = FASTPIXEL_Excludes::get_instance();
            $is_exclusion = $excludes->check_is_exclusion($url->get_url());
            if ($is_exclusion) {
                FASTPIXEL_Debug::log('Common Exclude and Meta Validation: Stopped. Page is excluded from cache.');
                //trying to delete existing files if page is exclusion and for some reason cached page exists, or it was created before exclusion was added
                $this->functions->delete_cached_files($url->get_url_path());
                return false;
            }
            //checking for meta file
            $meta_path = $this->cache_dir . DIRECTORY_SEPARATOR . $url->get_url_path() . 'meta';
            FASTPIXEL_Debug::log('Common Exclude and Meta Validation: checking URL meta', $meta_path);
            if (file_exists($meta_path)) {
                /*
                 * can't use here wordpress native functions(WP_Filesystem) because this function fires early in advanced-cache.php
                 */
                // phpcs:ignore WordPress.WP.AlternativeFunctions.file_get_contents_file_get_contents -- none available before WordPress is loaded.
                $meta = json_decode(file_get_contents($meta_path), true); //phpcs:ignore
                //need to check parent for pagination page
                if (preg_match('/page\/\d+/', $url->get_url())) {
                    $parent_url_path = preg_replace('/page.+$/i', '', $url->get_url_path());
                    $parent_meta = $this->functions->check_post_cache_status($parent_url_path);
                    //because can't invalidate paginated page, need to set parent's invalidation time
                    $meta['invalidated_time'] = $parent_meta['local_invalidation_time'];
                }

                //extra check for pages with params
                if (strpos($url->get_url(), '?') > 0) {
                    $url_path = preg_replace('/_[^\?]+$/i', '', $url->get_url_path());
                    $parent_status = $this->functions->check_post_cache_status($url_path);
                    if ((isset($meta['have_cache']) && $meta['have_cache']) 
                        && $parent_status['local_invalidation_time'] > $meta['local_invalidation_time'] &&
                        $meta['html_created_time'] < $parent_status['local_invalidation_time']
                    ) {
                        $meta['need_cache'] = true;
                        $meta['local_invalidation_time'] = $parent_status['local_invalidation_time'];
                    }
                }
            } else {
                //returning true if no meta file present, nothing to validate
                return true;
            }

            //check for last request time and prevent new request
            if ( isset($meta) &&
                (isset($meta['invalidated_time']) && $meta['invalidated_time'] <= $meta['cache_request_time']) &&
                (isset($meta['cache_request_time']) && time() < $meta['cache_request_time'] + $this->request_wait_time)
            ) {
                FASTPIXEL_Debug::log('Common Exclude and Meta Validation: Stopped, need to wait ' . $this->request_wait_time . ' secs until next retry');
                return false;
            }
            return true;
        }

        public function cache_request($reset_url)
        {
            if (empty($reset_url) || preg_match('/https?:\/\/\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}/i', $reset_url)) {
                return false;
            } else {
                if (!preg_match('/http/i', $reset_url)) {
                    $reset_url = (is_ssl() ? 'https://' : 'http://') . $reset_url;
                }
            }
            if (!defined('FASTPIXEL_REST_URL') || empty(FASTPIXEL_REST_URL)) {
                FASTPIXEL_DEBUG::log('CACHE REQUEST Error: FASTPIXEL_REST_URL is not defined or is empty');
                return false;
            }
            //comparing reset url domain with rest domain
            $compare_reset_url_domain = wp_parse_url($reset_url, PHP_URL_HOST);
            $compare_rest_url_domain = wp_parse_url(FASTPIXEL_REST_URL, PHP_URL_HOST);

            if (!preg_match('/^'.$compare_rest_url_domain.'/i', $compare_reset_url_domain)) {
                FASTPIXEL_DEBUG::log('CACHE REQUEST Error: $reset_url not match FASTPIXEL_REST_URL, request stopped');
                return false;
            }
            if (!defined('FASTPIXEL_API_URL') || empty(FASTPIXEL_API_URL)) {
                FASTPIXEL_DEBUG::log('CACHE REQUEST Error: FASTPIXEL_API_URL is not defined or is empty');
                return false;
            }

            $reset_url = apply_filters('fastpixel_cache_url_before_request', $reset_url);
            $params = [
                'url'         => $reset_url,
                'postbackUrl' => FASTPIXEL_REST_URL,
                'settings' => [
                    'modules' => []
                ]
            ];
            // phpcs:ignore WordPressVIPMinimum.Performance.WPQueryParams.PostNotIn_exclude -- data is not connected to wordpress query.
            $scriptrewrite = [
                // false to disable rewrite completely
                // do not include wp-meteor in the page disabled
                'enabled'  => false,
                'settings' => [
                    // optional
                    'exclude' => [ //phpcs:ignore
                        // optional
                        'strings' => [],
                        // optional, ["string1", "string2"]
                        'regexps' => [],
                        // optional, [{type => "RegExp", source => "domain\\.com"}, flags =>  "i"}, {type => "RegExp", source => "domain2\\.com"}, flags =>  "i"}]
                    ],
                    'eventHandlersRewrite' => false, // this is to match wp-meteor:0.1.39 to fpo-javascript-rewrite:0.1.40
                ]
            ];
            //1 => 'optimize', 2 => 'delaycritical', 3 => 'donotoptimize'
            $js_optimize = $this->functions->get_option('fastpixel_javascript_optimization', 1);
            $scriptrewrite['injectBootstrapScripts'] = true;
            if (in_array($js_optimize, [1,2])) {
                $scriptrewrite['enabled'] = true;
                if ($js_optimize == 1) {
                    $scriptrewrite['settings']['rdelay'] = 0;
                } else {
                    $scriptrewrite['settings']['rdelay'] = 86400 * 1000;
                }
                $js_excludes = $this->functions->get_option('fastpixel_javascript_excludes');
                $js_excludes_array = explode(chr(13), $js_excludes);
                $js_excludes_regexp = $this->functions->get_option('fastpixel_javascript_excludes_regexp');
                $js_excludes_regexp_array = explode(chr(13), $js_excludes_regexp);
                if (!empty($js_excludes_array) && !empty($js_excludes_array[0])) {
                    $scriptrewrite['settings']['exclude']['strings'] = $js_excludes_array;
                }
                if (!empty($js_excludes_regexp)) {
                    foreach($js_excludes_regexp_array as $regexp_exclude) {
                        $scriptrewrite['settings']['exclude']['regexps'][] = (object)['type' => 'RegExp', 'source' => $regexp_exclude, 'flags' => 'i'];
                    }
                }

                $exclude_gdpr = $this->functions->get_option('fastpixel_javascript_exclude_gdpr', 1);
                $scriptrewrite['settings']['excludeGDPR'] = $exclude_gdpr ? true : false;
            }
            $params['settings']['modules']['ScriptRewrite'] = $scriptrewrite;
            //images optimization
            $quality_params = array(1 => 'lossy', 2 => 'glossy', 3 => 'lossless');
            $images_optimization = $this->functions->get_option('fastpixel_images_optimization', 1); //default lossy
            $params['settings']['imageQuality'] = $quality_params[$images_optimization];
            //crop images
            $crop = $this->functions->get_option('fastpixel_images_crop', true); // true/false, default true
            $params['settings']['cropImages'] = (bool)$crop;
            //force images width/height attributes
            $force_dimensions = $this->functions->get_option('fastpixel_force_image_dimensions', false); // true/false, default false
            $params['settings']['modules']['ImageRewrite'] = [
                'enabled'  => true,
                'settings' => [
                    'forceImageDimensions' => (bool)$force_dimensions
                ]
            ];
            //fonts params
            $reduced_fonts = [
                'settings' => [
                    'soft' => false, //default value
                ]
            ];
            $force_soft_fonts = $this->functions->get_option('fastpixel_fonts_soft', false); // true/false, default false
            if ($force_soft_fonts) {
                $reduced_fonts['settings']['soft'] = true;
            }
            $params['settings']['modules']['ReducedFonts'] = $reduced_fonts;
            //doing request
            return $this->api_request(FASTPIXEL_API_URL, $params);
        }

        function api_request($api_url, $request_data) {
            $api_key = $this->functions->get_option('fastpixel_api_key');
            if (empty($api_key)) {
                FASTPIXEL_DEBUG::log('API REQUEST Error: Empty API KEY ', $api_key);
                return false;
            }
            //adding plugin version to all requests
            $request_data['plugin_version'] = FASTPIXEL_VERSION;
            $auth_api_key = base64_encode($api_key . ":");
            if (defined('FASTPIXEL_USE_SK') && FASTPIXEL_USE_SK === true) {
                $request_data['siteKey'] = $auth_api_key;
            }

            $connection_timeout = 5;
            if (function_exists('wp_remote_post')) {
                FASTPIXEL_DEBUG::log('API REQUEST : USING wp_remote_post');
                $args = array(
                    'timeout'     => $connection_timeout,
                    'sslverify'   => false,
                    'data_format' => 'body',
                    'headers'     => [
                        'Content-Type'  => 'application/json'
                    ],
                    'body'        => wp_json_encode($request_data),
                );
                if (!defined('FASTPIXEL_USE_SK') || FASTPIXEL_USE_SK == false) {
                    $args['headers'][] = ['Authorization' => 'Basic ' . $auth_api_key];
                }
                FASTPIXEL_DEBUG::log('WP_REMOTE_POST Request Params', $args);
                $response = wp_remote_post($api_url, $args);
                if (is_wp_error($response)) {
                    FASTPIXEL_DEBUG::log('WP_REMOTE_POST Response Error : ', $response);
                    return false;
                }
                FASTPIXEL_DEBUG::log('WP_REMOTE_POST Response', $response['response']);
                //validating server response
                if (!$this->default_handle_api_response($response['response']['code'], $response['body'], $request_data['url'])) {
                    return false;
                }
                return true;
            } else {
                FASTPIXEL_Debug::log('API REQUEST: function_exists(\'wp_remote_post\')', function_exists('wp_remote_post'));
            }
        }

        protected function default_handle_api_response($http_code = false, $response_body = false, $url = false) {
            if (empty($http_code)) {
                return false;
            }
            $body = json_decode($response_body, true);
            //checking if response is OK
            if ($http_code == 200) {
                //checking response body for 'queued' status
                if (isset($body['status']) && $body['status'] == 'queued') {
                    //removing limit reached message if it was set
                    if (defined('FASTPIXEL_FREE_LIMIT_REACHED') && FASTPIXEL_FREE_LIMIT_REACHED) {
                        $this->functions->update_option('fastpixel_free_limit_reached', false);
                    }
                    return true;
                } else {
                    FASTPIXEL_DEBUG::log('Default Handle Api Response: "Queued" is not present in response');
                }
            } else {
                //handling limit error
                if ($http_code == 429) {
                    $this->functions->update_option('fastpixel_free_limit_reached', true);
                    return false;
                }
                //handling error response, temporary disabling
                if (isset($body['error']) && !empty($body['error'])) {
                    if (!empty($url)) {
                        if ($this->functions->error_file($url, 'add', $body)) {
                            return false;
                        }
                    }
                }
            }
            return false;
        }

        public function get_request_wait_time() {
            return $this->request_wait_time;
        }

        //function that check api endpoints
        public function check_endpoints()
        {
            $api_key = $this->functions->get_option('fastpixel_api_key');
            if (empty($api_key)) {
                return false;
            }
            if (file_exists(FASTPIXEL_PLUGIN_DIR . DIRECTORY_SEPARATOR . 'endpoint.php')) {
                $last_request_time = filemtime(FASTPIXEL_PLUGIN_DIR . DIRECTORY_SEPARATOR . 'endpoint.php');
                if (time() < $last_request_time + $this->request_wait_time) {
                    FASTPIXEL_DEBUG::log('Detect Endpoints: Skipping request, need to wait ' . $this->request_wait_time . ' seconds. ');
                    return false;
                }
            }
            $auth_api_key = base64_encode($api_key . ":");

            if (function_exists('wp_remote_post')) {
                $request_data = ['url' => get_home_url()];
                $request_headers = [
                    'Content-Type'  => 'application/json'
                ];
                if (defined(FASTPIXEL_USE_SK) && FASTPIXEL_USE_SK) {
                    $request_data['siteKey'] = $auth_api_key;
                } else {
                    $request_headers['Authorization'] = 'Basic ' . $auth_api_key;
                }
                $args = array(
                    'timeout'     => 15,
                    'sslverify'   => false,
                    'data_format' => 'body',
                    'headers'     => $request_headers,
                    'body'        => wp_json_encode($request_data),
                );
                $response = wp_remote_post(FASTPIXEL_API_HOST.'/api/v1/detect_endpoints', $args);
                if (is_wp_error($response)) {
                    return false;
                } else if ($response['response']['code'] != 200) {
                    return false;
                }
                $body = json_decode($response['body'], true);
                $notices = FASTPIXEL_Notices::get_instance();
                if (is_array($body) && !empty($body)) {
                    $this->functions->update_ac_file($body[0]);
                    $notices->add_flash_notice(esc_html__('Endpoint updated', 'fastpixel-website-accelerator'), 'success');
                    return true;
                }
                /* translators: %s is used for support link */
                $notices->add_flash_notice(sprintf(esc_html__('Automatic endpoint fix failed, please contact %s.', 'fastpixel-website-accelerator'), '<a href=\'https://fastpixel.io/#contact\'>'. esc_html__('support', 'fastpixel-website-accelerator') . '</a>'), 'error', true);
                return false;
            }
        }

        public function load_modules()
        {
            if (!class_exists('FASTPIXEL\FASTPIXEL_Module')) {
                include_once(FASTPIXEL_PLUGIN_DIR . DIRECTORY_SEPARATOR . 'inc' . DIRECTORY_SEPARATOR . 'module.php');
            }
            if (file_exists(FASTPIXEL_PLUGIN_DIR . DIRECTORY_SEPARATOR . 'inc' . DIRECTORY_SEPARATOR . 'modules')) {
                if ($handle = opendir(FASTPIXEL_PLUGIN_DIR . DIRECTORY_SEPARATOR . 'inc' . DIRECTORY_SEPARATOR . 'modules')) {
                    while (false !== ($entry = readdir($handle))) {
                        if (!in_array($entry, ['.', '..'])) {
                            try {
                                include_once(FASTPIXEL_PLUGIN_DIR . DIRECTORY_SEPARATOR . 'inc' . DIRECTORY_SEPARATOR . 'modules' . DIRECTORY_SEPARATOR . $entry);
                            } catch (\Exception $e) {
                                FASTPIXEL_Debug::log('Exception message -> ', $e->getMessage());
                            }
                        }
                    }
                    closedir($handle);
                }
            }
        }

        public function display_cached_for_logged() {
            global $wp_styles, $wp_scripts;
            //need to split page into 2 parts
            $this->buffer;
            $regex = '/<\s?\/\s?body.*/s';
            preg_match($regex, $this->buffer, $matches);
            if (isset($matches[0]) && !empty($matches[0])) {
                $buffer = preg_replace($regex, '', $this->buffer);
                /*
                 * displaying main part(till closing body tag) of .html file, no need to escape output
                 */
                // phpcs:ignore WordPress.Security.EscapeOutput.OutputNotEscaped -- no need to escape.
                echo $buffer; //phpcs:ignore
                /*
                 * inserting admin styles and scripts to display adminbar
                 */
                foreach ($wp_styles->queue as $handle) {
                    wp_print_styles($handle);
                }
                foreach ($wp_scripts->queue as $handle) {
                    wp_print_scripts($handle);
                }
                wp_admin_bar_render();
                /* 
                 * displaying closing body tag, no need to escape output
                 */
                // phpcs:ignore WordPress.Security.EscapeOutput.OutputNotEscaped -- no need to escape.
                echo $matches[0]; //phpcs:ignore
            }
            exit;
        }
    }

    new FASTPIXEL_Cache();
}