<?php
namespace FASTPIXEL;
use WP_REST_Request;
use WP_REST_Response;
use WP_Error;
defined('ABSPATH') || exit;

if (!class_exists('FASTPIXEL\FASTPIXEL_Rest_Api')) {
    class FASTPIXEL_Rest_Api
    {
        public $functions;
        public static $instance;

        public function __construct()
        {
            self::$instance = $this;
            $this->functions = FASTPIXEL_functions::get_instance();
            if (defined('FASTPIXEL_SHORTINIT') && FASTPIXEL_SHORTINIT == true) {
                return;
            }
            register_rest_route(FASTPIXEL_TEXTDOMAIN . '/v1', '/update', 
                array(
                    'methods'             => 'POST',
                    'callback'            => [$this, 'check_request'],
                    'permission_callback' => '__return_true',
                )
            );
            register_rest_route(FASTPIXEL_TEXTDOMAIN . '/v1', '/version', 
                array(
                    'methods'             => 'GET',
                    'callback'            => [$this, 'version'],
                    'permission_callback' => '__return_true',
                )
            );
        }

        public static function get_instance()
        {
            if (empty(self::$instance)) {
                self::$instance = new FASTPIXEL_Rest_Api();
            }
            return self::$instance;
        }

        function save_files($parameters) // $url, $html, $headers, $css
        {
            //exclusion check moved here to have less duplicate code
            $excludes = FASTPIXEL_Excludes::get_instance();
            $is_exclusion = $excludes->check_is_exclusion($parameters['url']);
            $cache_dir = $this->functions->get_cache_dir();
            $url = new FASTPIXEL_Url(false, $parameters['url']);
            if ($is_exclusion) {
                //trying to delete existing files if page is exclusion and for some reason cached page exists
                $this->functions->delete_cached_files($cache_dir . DIRECTORY_SEPARATOR . $url->get_url_path());
                FASTPIXEL_Debug::log('WP REST API RESPONSE 400: page is excluded from cache');
                if (defined('FASTPIXEL_SHORTINIT') && FASTPIXEL_SHORTINIT == true) {
                    return false;
                } else {
                    return new WP_REST_Response(['status' => 400, 'response' => 'Bad Request', 'body_response' => 'Page is excluded from cache'], 400);
                }
            }

            $path = $this->functions->check_path($parameters['url']);
            if (!$path) {
                FASTPIXEL_Debug::log('REST API: skipping files save because no path returned');
                return false;
            }
            $modified_time = time(); // Make sure modified time is consistent.

            //initializing filesystem
            global $wp_filesystem;
            if (empty($wp_filesystem)) {
                require_once(ABSPATH . '/wp-admin/includes/file.php');
                WP_Filesystem();
            }
            // Save the response body.
            if (!$wp_filesystem->put_contents($path . DIRECTORY_SEPARATOR . 'index.html', $parameters['html'])) {
                FASTPIXEL_Debug::log('REST API: error occured while putting HTML file content to disk');
            }

            // Save the resonse headers.
            if (!$wp_filesystem->put_contents($path . DIRECTORY_SEPARATOR . 'headers.json', wp_json_encode(isset($parameters['headers']) ? $parameters['headers'] : null))){
                FASTPIXEL_Debug::log('REST API: error occured while putting headers file content to disk');
            }
            if (isset($parameters['css']) && !empty($parameters['css'])) {
                if (!$wp_filesystem->put_contents($path . DIRECTORY_SEPARATOR . 'style.css', $parameters['css'])) {
                    FASTPIXEL_Debug::log('REST API: error occured while putting css file content to disk');
                }
            }

            $wp_filesystem->touch($path . DIRECTORY_SEPARATOR . 'index.html', $modified_time);
            //need to remove error if it was stored
            $this->functions->error_file($parameters['url'], 'delete');

            //need to update request time if it was deleted previously to display page as cached, instead page will be displayed as stale
            $meta_file = $cache_dir . DIRECTORY_SEPARATOR . $url->get_url_path() . DIRECTORY_SEPARATOR . 'meta';
            if (!file_exists($meta_file)) {
                //adding meta info
                $meta = ['invalidated_time' => false, 'cache_request_time' => $modified_time - 1];
                if (!$wp_filesystem->put_contents($meta_file, wp_json_encode($meta))) {
                    FASTPIXEL_Debug::log('REST API: error occured while saving meta file');
                }
            }

            do_action('fastpixel_cache_files_saved', $parameters['url']);
            return true;
        }

        function check_request(WP_REST_Request $request)
        {
            FASTPIXEL_DEBUG::log('WP REST API Request Received');
            //getting request params first, added support for old WP
            if ((method_exists($request, 'is_json_content_type') && !$request->is_json_content_type()) ||
                $request->get_header('content_type') != 'application/json') {
                return new WP_REST_Response(['status' => 415, 'response' => 'Unsupported Media Type', 'body_response' => 'Unsupported Media Type'], 415);
            }
            $parameters = $request->get_json_params();
            //checking for authorization key
            if ((!isset($parameters['siteKey']) || empty($parameters['siteKey'])) && (empty($request->get_header('authorization')))) {
                FASTPIXEL_DEBUG::log('WP REST API RESPONSE 401');
                return new WP_REST_Response(['status' => 401, 'response' => 'Not Authorized'], 401);
            } 
            //checking for posted authorization sitekey first, then checking authorization headers
            if (isset($parameters['siteKey']) && !empty($parameters['siteKey'])) {
                $site_key_decoded = base64_decode($parameters['siteKey']);
                $request_key = substr($site_key_decoded, 0, strlen($site_key_decoded) - 1);
            } else {
                $request_key = isset($_SERVER['PHP_AUTH_USER']) && !empty($_SERVER['PHP_AUTH_USER']) ? $_SERVER['PHP_AUTH_USER'] : null;
            }
            //getting api key
            $api_key = $this->functions->get_option('fastpixel_api_key');
            if ($request_key !== $api_key) {
                FASTPIXEL_DEBUG::log('WP REST API RESPONSE 401');
                return new WP_REST_Response(['status' => 401, 'response' => 'Not Authorized'], 403);
            }
            if (empty($parameters['url'])) {
                FASTPIXEL_DEBUG::log('WP REST API RESPONSE 400: url parameter is missing');
                return new WP_REST_Response(['status' => 400, 'response' => 'Bad Request', 'body_response' => 'url parameter is missing'], 400);
            }
            try {
                $url = wp_parse_url($parameters['url']);
                if (empty($url['scheme'])) {
                    throw new Exception("No scheme");
                }
                // TODO: temporarily disabled host check
                // if ($url['host'] !== $_SERVER['HTTP_HOST']) {
                //     throw new Exception("Invalid host");
                // }
            } catch (Exception $e) {
                FASTPIXEL_DEBUG::log('WP REST API RESPONSE 400: invalid url parameter');
                return new WP_REST_Response(['status' => 400, 'response' => 'Bad Request', 'body_response' => 'invalid url parameter: ' . $e->getMessage()], 400);
            }
            //checking for error and writing error file if exists
            if (isset($parameters['error']) && !empty($parameters['error'])) {
                $this->functions->error_file($parameters['url'], 'add', ['error' => $parameters['error']]);
                return new WP_REST_Response(['status' => 200, 'response' => 'ok', 'body_response' => 'Error saved'], 200);
            }
            if (empty($parameters['html'])) {
                FASTPIXEL_DEBUG::log('WP REST API RESPONSE 400: html parameter is missing');
                return new WP_REST_Response(['status' => 400, 'response' => 'Bad Request', 'body_response' => 'html parameter is missing'], 400);
            }
            FASTPIXEL_DEBUG::log('WP REST API RESPONSE 200 OK');
            if ($this->save_files($parameters)) {
                return new WP_REST_Response(['status' => 'ok'], 200);
            } else {
                return new WP_REST_Response(['status' => 400, 'response' => 'Bad Request', 'body_response' => 'Request data processing failed'], 400);
            }
        }

        public function version() {
            return new WP_REST_Response(['version' => FASTPIXEL_VERSION], 200);
        }
    }
    new FASTPIXEL_Rest_Api();
}